# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import keyword
import logging
import re
import time

from ..source.base import DataSource, DataSourceBase, NoEntry
from .utils import reload_on_change

logger = logging.getLogger("intake")


class Catalog(DataSource):
    """Manages a hierarchy of data sources as a collective unit.

    A catalog is a set of available data sources for an individual
    entity (remote server, local  file, or a local
    directory of files). This can be expanded to include a
    collection of subcatalogs, which are then managed as a single unit.

    A catalog is created with a single URI or a group of URIs. A URI can
    either be a URL or a file path.

    Each catalog in the hierarchy is responsible for caching the most recent
    refresh time to prevent overeager queries.

    Attributes
    ----------
    metadata : dict
        Arbitrary information to carry along with the data source specs.
    """

    # emulate a DataSource
    container = "catalog"
    name = "catalog"
    auth = None

    def __init__(
        self,
        entries=None,
        name=None,
        description=None,
        metadata=None,
        ttl=60,
        getenv=True,
        getshell=True,
        persist_mode="default",
        storage_options=None,
        user_parameters=None,
    ):
        """
        Parameters
        ----------
        entries : dict, optional
            Mapping of {name: entry}
        name : str, optional
            Unique identifier for catalog. This takes precedence over whatever
            is stated in the cat file itself. Defaults to None.
        description : str, optional
            Description of the catalog. This takes precedence over whatever
            is stated in the cat file itself. Defaults to None.
        metadata: dict
            Additional information about this data
        ttl : float, optional
            Lifespan (time to live) of cached modification time. Units are in
            seconds. Defaults to 1.
        getenv: bool
            Can parameter default fields take values from the environment
        getshell: bool
            Can parameter default fields run shell commands
        persist_mode: ['always', 'default', 'never']
            Defines the use of persisted sources: if 'always', will use a
            persisted version of a data source, if it exists, if 'never' will
            always use the original source. If 'default', persisted sources
            will be used if they have not expired, and re-persisted and used
            if they have.
        storage_options : dict
            If using a URL beginning with 'intake://' (remote Intake server),
            parameters to pass to requests when issuing http commands; otherwise
            parameters to pass to remote backend file-system. Ignored for
            normal local files.
        """
        super(Catalog, self).__init__()
        self.name = name
        self.description = description
        self.metadata = metadata or {}
        self.ttl = ttl
        self.getenv = getenv
        self.getshell = getshell
        self.storage_options = storage_options
        if isinstance(user_parameters, dict) and user_parameters:
            from .local import UserParameter

            self.user_parameters = {name: (UserParameter(name=name, **up) if isinstance(up, dict) else up) for name, up in user_parameters.items()}
        elif isinstance(user_parameters, (list, tuple)):
            self.user_parameters = {up["name"]: up for up in user_parameters}
        else:
            self.user_parameters = {}
        if persist_mode not in ["always", "never", "default"]:
            # should be True, False, None ?
            raise ValueError("Persist mode (%s) not understood" % persist_mode)
        self.pmode = persist_mode

        if entries and isinstance(entries, str):
            raise ValueError(
                "The class intake.Catalog does not accept a string for "
                "`entries`\n"
                "Did you mean to use `intake.open_catalog`? Note that in "
                "versions of intake <=0.5.4 `intake.Catalog` was an "
                "alias for `intake.open_catalog`. It is now the intake base "
                "Catalog class."
            )
        self.updated = time.time()
        self._entries = entries if entries is not None else self._make_entries_container()
        self.force_reload()

    @classmethod
    def from_dict(cls, entries, **kwargs):
        """
        Create Catalog from the given set of entries

        Parameters
        ----------
        entries : dict-like
            A mapping of name:entry which supports dict-like functionality,
            e.g., is derived from ``collections.abc.Mapping``.
        kwargs : passed on the constructor
            Things like metadata, name; see ``__init__``.

        Returns
        -------
        Catalog instance
        """
        cat = cls(**kwargs)
        cat._entries = entries
        return cat

    @property
    def kwargs(self):
        return dict(name=self.name, ttl=self.ttl)

    def _make_entries_container(self):
        """Subclasses may override this to return some other dict-like.

        See RemoteCatalog below for the motivating example for this hook. This
        is typically useful for large Catalogs backed by dynamic resources such
        as databases.

        The object returned by this method must implement:

        * ``__iter__()`` -> an iterator of entry names
        * ``__getitem__(key)`` -> an Entry
        * ``items()`` -> an iterator of ``(key, Entry)`` pairs

        For best performance the object should also implement:

        * ``__len__()`` -> int
        * ``__contains__(key)`` -> boolean

        In ``__len__`` or ``__contains__`` are not implemented, intake will
        fall back on iterating through the entire catalog to compute its length
        or check for containment, which may be expensive on large catalogs.
        """
        return {}

    def _load(self):
        """Override this: load catalog entries"""
        pass

    def force_reload(self):
        """Imperative reload data now"""
        self.updated = time.time()
        self._load()

    def reload(self):
        """Reload catalog if sufficient time has passed"""
        if (self.ttl is not None) and (time.time() - self.updated > self.ttl):
            self.force_reload()

    @property
    def version(self):
        # default version for pre-v1 files
        return self.metadata.get("version", 1)

    @reload_on_change
    def search(self, text, depth=2):
        import copy

        words = text.lower().split()
        entries = {k: copy.copy(v) for k, v in self.walk(depth=depth).items() if any(word in str(v.describe().values()).lower() for word in words)}
        cat = Catalog.from_dict(
            entries,
            name=self.name + "_search",
            ttl=self.ttl,
            getenv=self.getenv,
            getshell=self.getshell,
            metadata=(self.metadata or {}).copy(),
            storage_options=self.storage_options,
            user_parameters=self.user_parameters.copy(),
        )
        cat.metadata["search"] = {"text": text, "upstream": self.name}
        cat.cat = self
        for e in entries.values():
            e._catalog = cat
        return cat

    def filter(self, func):
        """
        Create a Catalog of a subset of entries based on a condition

        .. warning ::

           This function operates on CatalogEntry objects not DataSource
           objects.

        .. note ::

            Note that, whatever specific class this is performed on,
            the return instance is a Catalog. The entries are passed
            unmodified, so they will still reference the original catalog
            instance and include its details such as directory,.

        Parameters
        ----------
        func : function
            This should take a CatalogEntry and return True or False. Those
            items returning True will be included in the new Catalog, with the
            same entry names

        Returns
        -------
        Catalog
           New catalog with Entries that still refer to their parents
        """
        return Catalog.from_dict({key: entry for key, entry in self._entries.items() if func(entry)})

    @reload_on_change
    def walk(self, sofar=None, prefix=None, depth=2):
        """Get all entries in this catalog and sub-catalogs

        Parameters
        ----------
        sofar: dict or None
            Within recursion, use this dict for output
        prefix: list of str or None
            Names of levels already visited
        depth: int
            Number of levels to descend; needed to truncate circular references
            and for cleaner output

        Returns
        -------
        Dict where the keys are the entry names in dotted syntax, and the
        values are entry instances.
        """
        out = sofar if sofar is not None else {}
        prefix = [] if prefix is None else prefix
        for name, item in self._entries.items():
            if item._container == "catalog" and depth > 1:
                # recurse with default open parameters
                try:
                    item().walk(out, prefix + [name], depth - 1)
                except Exception as e:
                    print(e)
                    pass  # ignore inability to descend
            n = ".".join(prefix + [name])
            out[n] = item
        return out

    def items(self):
        """Get an iterator over (key, source) tuples for the catalog entries."""
        for name, entry in self._get_entries().items():
            yield name, entry()

    def values(self):
        """Get an iterator over the sources for catalog entries."""
        for entry in self._get_entries().values():
            yield entry()

    def serialize(self):
        """
        Produce YAML version of this catalog.

        Note that this is not the same as ``.yaml()``, which produces a YAML
        block referring to this catalog.
        """
        import yaml

        output = {"metadata": self.metadata, "sources": {}, "name": self.name, "description": self.description}
        for key, entry in self._entries.items():
            kw = entry._captured_init_kwargs.copy()
            kw.pop("catalog", None)
            kw["parameters"] = {k.name: k.__getstate__()["kwargs"] for k in kw.get("parameters", [])}
            try:
                if issubclass(kw["driver"], DataSourceBase):
                    kw["driver"] = ".".join([kw["driver"].__module__, kw["driver"].__name__])
            except TypeError:
                pass  # ignore exception for a string input
            output["sources"][key] = kw
        return yaml.dump(output)

    def save(self, url, storage_options=None):
        """
        Output this catalog to a file as YAML

        Parameters
        ----------
        url : str
            Location to save to, perhaps remote
        storage_options : dict
            Extra arguments for the file-system
        """
        from fsspec import open_files

        with open_files([url], **(storage_options or {}), mode="wt")[0] as f:
            f.write(self.serialize())

    @reload_on_change
    def _get_entry(self, name):
        entry = self._entries[name]
        entry._catalog = self
        entry._pmode = self.pmode

        up_names = set((up["name"] if isinstance(up, dict) else up.name) for up in entry._user_parameters)
        ups = [up for name, up in self.user_parameters.items() if name not in up_names]
        entry._user_parameters = ups + (entry._user_parameters or [])
        return entry()

    def configure_new(self, **kwargs):
        from .local import UserParameter

        ups = {}
        for k, v in kwargs.copy().items():
            for up in self.user_parameters.values():
                if isinstance(up, dict):
                    if k == up["name"]:
                        kw = up.copy()
                        kw["default"] = v
                        ups[k] = UserParameter(**kw)
                        kwargs.pop(k)
                else:
                    if k == up.name:
                        kw = up._captured_init_kwargs.copy()
                        kw["default"] = v
                        kw["name"] = k
                        ups[k] = UserParameter(**kw)
                        kwargs.pop(k)

        new = super().configure_new(**kwargs)
        new.user_parameters.update(ups)
        return new

    __call__ = get = configure_new

    @reload_on_change
    def _get_entries(self):
        return self._entries

    def __iter__(self):
        """Return an iterator over catalog entry names."""
        return iter(self._get_entries())

    def keys(self):
        """Entry names in this catalog as an iterator (alias for __iter__)"""
        return iter(self)

    def __len__(self):
        return len(self._get_entries())

    def __contains__(self, key):
        # Avoid iterating through all entries.
        return key in self._get_entries()  # triggers reload_on_change

    def __dir__(self):
        # Include tab-completable entries and normal attributes.
        return [
            entry for entry in self if re.match("[_A-Za-z][_a-zA-Z0-9]*$", entry) and not keyword.iskeyword(entry)  # valid Python identifer
        ] + list(  # not a Python keyword
            self.__dict__.keys()
        )

    def _ipython_key_completions_(self):
        return list(self)

    def __repr__(self):
        return "<Intake catalog: %s>" % self.name

    def __getattr__(self, item):
        # we need this special case here because the (deprecated) entry
        # property on the base class
        if item == "entry":
            raise NoEntry("Source was not made from a catalog entry")
        if not item.startswith("_"):
            # Fall back to __getitem__.
            try:
                return self[item]  # triggers reload_on_change
            except KeyError as e:
                raise AttributeError(item) from e
        raise AttributeError(item)

    def __setitem__(self, key, entry):
        """Add entry to catalog

        This relies on the `_entries` attribute being mutable, which it normally
        is. Note that if a catalog automatically reloads, any entry added here
        may be very transient

        Parameters
        ----------
        key : str
            Key to give the entry in the cat
        entry : CatalogEntry
            The entry to include (could be local, remote)
        """
        self._entries[key] = entry

    def pop(self, key):
        """Remove entry from catalog and return it

        This relies on the `_entries` attribute being mutable, which it normally
        is. Note that if a catalog automatically reloads, any entry removed here
        may soon reappear

        Parameters
        ----------
        key : str
            Key to give the entry in the cat
        """
        return self._entries.pop(key)

    def __getitem__(self, key):
        """Return a catalog entry by name.

        Can also use attribute syntax, like ``cat.entry_name``, or
        item lookup cat['non-python name']. This enables walking through
        nested directories with cat.name1.name2, cat['name1.name2'] *or*
        cat['name1', 'name2']
        """
        if not isinstance(key, list) and key in self:
            # triggers reload_on_change
            s = self._get_entry(key)
            if s.container == "catalog":
                s.name = key
                s.user_parameters.update(self.user_parameters.copy())
                return s
            return s
        if isinstance(key, str) and "." in key:
            key = key.split(".")
        if isinstance(key, list):
            parts = list(key)[:]
            prefix = ""
            while parts:
                bit = parts.pop(0)
                prefix = prefix + ("." if prefix else "") + bit
                if prefix in self._entries:
                    rest = ".".join(parts)
                    try:
                        out = self._entries[prefix][rest]
                        return out()
                    except KeyError:
                        # name conflict like "thing" and "think.oi", where it's
                        # the latter we are after
                        continue
        elif isinstance(key, tuple):
            out = self
            for part in key:
                out = out[part]
            return out()
        raise KeyError(key)

    def discover(self):
        return {"container": "catalog", "shape": None, "dtype": None, "metadata": self.metadata}

    def _close(self):
        # TODO: maybe close all entries?
        pass

    @property
    def gui(self):
        if not hasattr(self, "_gui"):
            from ..interface import output_notebook
            from ..interface.gui import GUI

            output_notebook()
            self._gui = GUI(self)
        else:
            self._gui.visible = True
        return self._gui



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2022, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
import hvplot
from packaging.version import Version

try:
    import xrviz
    from xrviz.dashboard import Dashboard as XRViz

    assert Version(xrviz.__version__) >= Version("0.1.1")
except ImportError:
    xrviz = False

import panel as pn

from ...catalog.local import LocalCatalogEntry, YAMLFileCatalog
from ..base import BaseView


class Event:
    def __init__(self, plot, **kwargs):
        self.object = plot
        for key, val in kwargs.items():
            setattr(self, key, val)


class Plots(BaseView):
    """
    Panel for displaying pre-defined plots from catalog.

    Parameters
    ----------
    source: intake catalog entry, or list of same
        source to describe in this object
    edit_callback: callback to alert that plot has been edited

    Attributes
    ----------
    has_plots: bool
        whether the source has plots defined
    options: list
        plots options defined on the source
    selected: str
        name of selected plot
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """

    select = None

    def __init__(self, source=None, **kwargs):
        self.custom = pn.widgets.Button(name="Create", width=70, align="center")
        self.source = source
        self.panel = pn.Column(name="Plot", width_policy="max", margin=0)
        self._behavior_callback = None
        self._callbacks = []
        super().__init__(**kwargs)

    def setup(self):
        self.instructions = pn.pane.Markdown("**Select Plot**", align="center", width=70)
        self.select = pn.widgets.Select(options=self.options, height=30, align="center", min_width=200)
        # Add spaces in front of each option to make it render correctly.
        # This is a bit of a hack to make a Select which only displays the down arrow.
        self.edit_options = pn.widgets.Select(options=["  Edit", "  Clone", "  Rename", "  Delete"], align="center", width=25, margin=(5, -10))
        self.pane = pn.pane.HoloViews(self._plot_object(self.selected), name="Plot")

        self.interact_label = pn.pane.Markdown("Name", align="center", width_policy="max", min_width=100)
        self.interact_name = pn.widgets.TextInput(placeholder="Name of new plot...")
        self.interact_cancel = pn.widgets.Button(name="Cancel", width=100)
        self.interact_save = pn.widgets.Button(name="Save", button_type="primary", width=100)

        self.watchers = [
            self.select.param.watch(self.plot_selected, ["value"]),
            self.custom.param.watch(self.interact, ["clicks"]),
            self.edit_options.param.watch(self.interact, ["value"]),
            self.interact_name.param.watch(self.name_changed, ["value_input"]),
            self.interact_cancel.param.watch(self.cancel, ["clicks"]),
            self.interact_save.param.watch(self.interact_action, ["clicks"]),
        ]
        self.alert = pn.pane.Alert(alert_type="danger")
        self.out = pn.Row(self.pane, name="Plot")

        self.row_select_plots = pn.Row(
            self.instructions,
            self.select,
            self.custom,
            self.edit_options,
        )
        self.row_dialog_buttons = pn.Row(
            self.interact_label,
            self.interact_name,
            self.interact_cancel,
            self.interact_save,
        )
        self.children = [
            pn.Column(
                self.row_select_plots,
                self.row_dialog_buttons,
                pn.Row(self.alert),
                self.out,
            )
        ]
        # Set initial visibility
        self.row_select_plots[-1].visible = False  # edit_options dropdown
        self.row_dialog_buttons.visible = False
        self.alert.visible = False

    @BaseView.source.setter
    def source(self, source):
        """When the source gets updated, update the options in
        the selector"""
        if source and isinstance(source, list):
            source = source[0]
        if isinstance(source, LocalCatalogEntry):
            source = source()
        BaseView.source.fset(self, source)
        if self.select:
            self.select.options = self.options
        if source and source.container == "dataframe":
            self.custom.disabled = False
        elif source and xrviz and source.container in ["xarray", "ndarray", "numpy"]:
            self.custom.disabled = False
        else:
            self.custom.disabled = True

    @property
    def has_plots(self):
        """Whether the source has plots defined"""
        return self.source is not None and len(self._source.plots) > 0

    @property
    def options(self):
        """Plots options defined on the source"""
        return (["None"] + self.source.plots) if self.source is not None else []

    @property
    def selected(self):
        """Name of selected plot"""
        return self.select.value if self.select is not None else None

    @selected.setter
    def selected(self, selected):
        """When plot is selected set, make sure widget stays upto date"""
        self.select.value = selected

    def watch(self, callback):
        self._callbacks.append(callback)

    def plot_selected(self, *events):
        for event in events:
            if event.name == "value":
                self.pane.object = self._plot_object(event.new)
            self.custom.name = "Create" if str(self.select.value) == "None" else "Edit"
            self.edit_options.visible = self.custom.name == "Edit"

    def name_changed(self, *events):
        for event in events:
            if event.name == "value_input":
                self.alert.visible = False
                # Empty name not allowed and name cannot already be in use
                if len(event.new) == 0:
                    self.interact_save.disabled = True
                elif event.new in self.options:
                    self.interact_save.disabled = True
                    self.alert.object = f'Name "{event.new}" already exists'
                    self.alert.visible = True
                else:
                    self.interact_save.disabled = False

    def cancel(self, _):
        self.pane.object = self._plot_object(self.selected)
        self.out[0] = self.pane
        self.row_select_plots.visible = True
        self.row_dialog_buttons.visible = False
        self._behavior_callback = None

    def interact_action(self, _):
        # Stop catalog from autoreloading
        self.source.cat.ttl = None
        if self._behavior_callback is not None:
            # Perform action
            event = self._behavior_callback()
            # Save catalog (if YAMLFileCatalog)
            if isinstance(self.source.cat, YAMLFileCatalog):
                self.source.cat.add(self.source)
            # Callbacks
            for cb in self._callbacks:
                cb(event)
        # Reset custom behavior callback
        self._behavior_callback = None
        # End action
        self.cancel(None)

    def interact(self, _):
        # Create/Edit/Clone/Rename/Delete was selected
        if self._behavior_callback is not None:
            # This is a short-circuit to allow edit options reset
            # without causing an infinite loop
            return
        if self.selected == "None":
            # Create
            self._behavior_callback = self._create
            self.interact_save.name = "Save"
            self.interact_save.disabled = True
            self.interact_label.object = "Name "
            self.interact_name.value = ""
            self.interact_name.visible = True
            viz = self.draw()
            self.out[0] = viz
        elif self.edit_options.value == "  Clone":
            self._behavior_callback = self._clone
            self.interact_save.name = "Clone"
            self.interact_save.disabled = True
            self.interact_label.object = f'Clone "**{self.selected}**" as '
            self.interact_name.value = ""
            self.interact_name.visible = True
        elif self.edit_options.value == "  Rename":
            self._behavior_callback = self._rename
            self.interact_save.name = "Rename"
            self.interact_save.disabled = True
            self.interact_label.object = f'Rename "**{self.selected}**" to '
            self.interact_name.value = ""
            self.interact_name.visible = True
        elif self.edit_options.value == "  Delete":
            self._behavior_callback = self._delete
            self.interact_save.name = "Delete"
            self.interact_save.disabled = False
            self.interact_label.object = f'Really delete "**{self.selected}**" ?'
            self.interact_name.visible = False
        elif self.edit_options.value == "  Edit":
            # Edit
            self._behavior_callback = self._edit
            self.interact_save.name = "Save"
            self.interact_save.disabled = False
            self.interact_label.object = f'Editing "**{self.selected}**"'
            self.interact_name.visible = False
            viz = self.draw()
            self.out[0] = viz
        else:
            raise ValueError(self.edit_options.value)
        # Update visibility of components
        self.row_dialog_buttons.visible = True
        self.row_select_plots.visible = False
        # Reset edit options selection (won't trigger
        # edit action because _behavior_callback is set)
        self.edit_options.value = "Edit"

    def draw(self):
        if self.selected == "None":
            kwargs = {"y": []}
        else:
            kwargs = self.source.metadata["plots"][self.selected]
        if self.source.container == "dataframe":
            df = self.source.to_dask()
            if df.npartitions == 1:
                df = df.compute()
            self._viz = viz = hvplot.explorer(df, **kwargs)
        elif self.source.container in ["xarray", "ndarray", "numpy"]:
            import xarray

            try:
                data = self.source.to_dask()
            except NotImplementedError:
                data = self.source.read()
            if not isinstance(data, (xarray.DataArray, xarray.Dataset)):
                data = xarray.DataArray(data)
            self._viz = XRViz(data, **kwargs)
            viz = self._viz.panel
        else:
            raise ValueError(f"Unhandled container type {self.source.container}")
        return viz

    def _plot_object(self, selected):
        if selected and str(selected) != "None":
            plot_method = getattr(self.source.plot, selected)
            self.out[0] = self.pane
            if plot_method:
                return plot_method()

    def _create(self):
        plot_name = self.interact_name.value
        # Add plot metadata to both DataSource and CatalogEntry
        self.source.metadata.setdefault("plots", {})[plot_name] = self._viz.settings()
        self.source.entry._metadata.setdefault("plots", {})[plot_name] = self._viz.settings()
        # Add new plot name to self.options
        self.select.options = self.options
        # Select new graph
        self.selected = plot_name
        return Event(self, kind="create", plot_name=plot_name)

    def _edit(self):
        # Update plot metadata for both DataSource and CatalogEntry
        self.source.metadata["plots"][self.selected] = self._viz.settings()
        self.source.entry._metadata["plots"][self.selected] = self._viz.settings()
        return Event(self, kind="edit", plot_name=self.selected)

    def _clone(self):
        plot_name = self.interact_name.value
        # Clone plot metadata for both DataSource and CatalogEntry
        self.source.metadata["plots"][plot_name] = self.source.metadata["plots"][self.selected].copy()
        self.source.entry._metadata["plots"][plot_name] = self.source.entry._metadata["plots"][self.selected].copy()
        # Add new plot name to self.options
        self.select.options = self.options
        # Select new graph
        self.selected = plot_name
        return Event(self, kind="clone", plot_name=plot_name, cloned_from=self.selected)

    def _rename(self):
        plot_name = self.interact_name.value
        # Rename plot metadata for both DataSource and CatalogEntry
        self.source.metadata["plots"][plot_name] = self.source.metadata["plots"].pop(self.selected)
        self.source.entry._metadata["plots"][plot_name] = self.source.entry._metadata["plots"].pop(self.selected)
        # Update available options in dropdown
        self.select.options = self.options
        # Select renamed graph
        self.selected = plot_name
        return Event(self, kind="rename", plot_name=plot_name, prev_name=self.selected)

    def _delete(self):
        # Delete plot metadata from both DataSource and CatalogEntry
        del self.source.metadata["plots"][self.selected]
        del self.source.entry._metadata["plots"][self.selected]
        # Update available options in dropdown
        self.select.options = self.options
        self.selected = "None"
        return Event(self, kind="delete", plot_name=self.selected)

    def __getstate__(self, include_source=True):
        """Serialize the current state of the object. Set include_source
        to False when using with another panel that will include source."""
        state = super().__getstate__(include_source)
        state.update(
            {
                "selected": self.selected,
            }
        )
        return state

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        super().__setstate__(state)
        if self.visible:
            self.selected = state.get("selected")
        return self



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
import collections
import inspect
import logging
import os
import warnings

import entrypoints
from fsspec import get_filesystem_class, open_files
from fsspec.core import split_protocol

from .. import __version__
from ..source import get_plugin_class, register_driver
from ..source.discovery import load_plugins_from_module
from ..utils import DictSerialiseMixin, classname, make_path_posix, yaml_load
from . import exceptions
from .base import Catalog, DataSource
from .entry import CatalogEntry
from .utils import COERCION_RULES, _has_catalog_dir, coerce, expand_defaults, merge_pars

logger = logging.getLogger("intake")


class UserParameter(DictSerialiseMixin):
    """
    A user-settable item that is passed to a DataSource upon instantiation.

    For string parameters, default may include special functions ``func(args)``,
    which *may* be expanded from environment variables or by executing a shell
    command.

    Parameters
    ----------
    name: str
        the key that appears in the DataSource argument strings
    description: str
        narrative text
    type: str
        one of list ``(COERSION_RULES)``
    default: type value
        same type as ``type``. It a str, may include special functions
        env, shell, client_env, client_shell.
    min, max: type value
        for validation of user input
    allowed: list of type
        for validation of user input
    """

    def __init__(self, name, description=None, type=None, default=None, min=None, max=None, allowed=None):
        self.name = name
        self.description = description
        self.type = type or __builtins__["type"](default).__name__
        self.min = min
        self.max = max
        self.allowed = allowed

        self._default = default
        try:
            self.default = coerce(self.type, default)
        except (ValueError, TypeError):
            self.default = None
        self.expanded_default = self.default

        if self.min:
            self.min = coerce(self.type, self.min)

        if self.max:
            self.max = coerce(self.type, self.max)

        if self.allowed and type != "mlist":
            self.allowed = [coerce(self.type, item) for item in self.allowed]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"

    __str__ = __repr__

    def describe(self):
        """Information about this parameter"""
        desc = {
            "name": self.name,
            "description": self.description,
            # the Parameter might not have a type at all
            "type": self.type or "unknown",
        }
        for attr in ["min", "max", "allowed", "default"]:
            v = getattr(self, attr)
            if v is not None:
                desc[attr] = v
        return desc

    def expand_defaults(self, client=False, getenv=True, getshell=True):
        """Compile env, client_env, shell and client_shell commands"""
        if not isinstance(self._default, str):
            self.expanded_default = self._default
        else:
            self.expanded_default = coerce(self.type, expand_defaults(self._default, client, getenv, getshell))

    def validate(self, value):
        """Does value meet parameter requirements?"""
        if self.type is not None:
            value = coerce(self.type, value)

        if self.type == "mlist":
            for v in value:
                if v not in self.allowed:
                    raise ValueError("Item %s not in allowed list", v)
            return value

        if self.min is not None and value < self.min:
            raise ValueError("%s=%s is less than %s" % (self.name, value, self.min))
        if self.max is not None and value > self.max:
            raise ValueError("%s=%s is greater than %s" % (self.name, value, self.max))
        if self.allowed is not None and value not in self.allowed:
            raise ValueError("%s=%s is not one of the allowed values: %s" % (self.name, value, ",".join(map(str, self.allowed))))

        return value


class LocalCatalogEntry(CatalogEntry):
    """A catalog entry on the local system"""

    def __init__(
        self, name, description, driver, direct_access=True, args={}, cache=[], parameters=[], metadata={}, catalog_dir="", getenv=True, getshell=True, catalog=None
    ):
        """

        Parameters
        ----------
        name: str
            How this entry is known, normally from its key in a YAML file, or
            if that is not provided then from name of file, or name of dir if
            file name is 'catalog.yaml' or 'catalog.yml'.
        description: str
            Brief text about the target source
        driver: str, list, dict or DataSource subclass
            The plugin(s) that can load this. Can be a simple name like "csv",
            which will be looked up in the registry, a fully-qualified class
            name ("package.mod.Class"), a list of these which would all work,
            a dictionary of the same with reasonable names, or an explicit
            class derived from DataSource.
        direct_access: bool
            Is the client allowed to attempt to reach this data
        args: dict
            Passed when instantiating the plugin DataSource
        parameters: list
            UserParameters that can be set
        metadata: dict
            Additional information about this data
        catalog_dir: str
            Location of the catalog, if known
        getenv: bool
            Can parameter default fields take values from the environment
        getshell: bool
            Can parameter default fields run shell commands
        catalog: bool
            Catalog object in which this entry belongs
        """
        self._name = name
        self._default_source = None
        self._description = description
        self._driver = driver
        self._direct_access = direct_access
        self._open_args = args
        self._cache = cache
        self._user_parameters = parameters
        self._metadata = metadata or {}
        self._catalog_dir = catalog_dir
        self._filesystem = None
        self._catalog = catalog
        if isinstance(driver, str):
            dr = get_plugin_class(driver)
            self._plugin = [dr] if dr is not None else []
            containers = set(p.container for p in self._plugin)
        elif isinstance(driver, list):
            self._plugin = [get_plugin_class(d) for d in driver]
            self._plugin = [p for p in self._plugin if p is not None]
            containers = set(p.container for p in self._plugin)
        elif isinstance(driver, dict):
            self._plugin = {d: get_plugin_class(driver[d]["class"]) for d in driver}
            self._plugin = {k: v for k, v in self._plugin.items() if v is not None}
            containers = set(p.container for p in self._plugin.values())
        elif inspect.isclass(driver) and issubclass(driver, DataSource):
            self._plugin = [driver]
            containers = {driver.container}
        else:
            raise TypeError("Driver was not a string, list, dict or DataSource:" " %s" % driver)
        if len(containers) > 1:
            # this is an error, because cat is poorly specified, even if other
            # plugins are OK
            raise ValueError("Plugins for a data source must have only one " "container.")
        if len(containers) == 0:
            # this is only debug, this single entry won't work, but cat is OK.
            # you get an error if you try to actually use this entry
            logger.debug("No plugins for entry: %s" % self.name)
            containers = [None]
        self._container = list(containers)[0]
        super(LocalCatalogEntry, self).__init__(getenv=getenv, getshell=getshell)

    @property
    def name(self):
        return self._name

    def describe(self):
        """Basic information about this entry"""
        if isinstance(self._plugin, list):
            pl = [p.name for p in self._plugin]
        elif isinstance(self._plugin, dict):
            pl = {k: classname(v) for k, v in self._plugin.items()}
        else:
            pl = self._plugin if isinstance(self._plugin, str) else self._plugin.name
        return {
            "name": self._name,
            "container": self._container,
            "plugin": pl,  # deprecated
            "driver": pl,
            "description": self._description,
            "direct_access": self._direct_access,
            "user_parameters": [u.describe() for u in self._user_parameters],
            "metadata": self._metadata,
            "args": self._open_args,
        }

    def _create_open_args(self, user_parameters):
        plugin = user_parameters.pop("plugin", None)

        md = self._metadata.copy() if self._metadata is not None else {}
        md["catalog_dir"] = self._catalog_dir
        if user_parameters.pop("cache", None) or self._cache:
            md["cache"] = user_parameters.pop("cache", None) or self._cache
        params = {
            "metadata": md,
            "CATALOG_DIR": self._catalog_dir,
        }
        params.update(self._open_args)
        if "storage_options" not in params and self._filesystem is not None and self._filesystem.storage_options and _has_catalog_dir(params):
            params["storage_options"] = self._filesystem.storage_options
        open_args = merge_pars(params, user_parameters, self._user_parameters, getshell=self.getshell, getenv=self.getenv, client=False)

        if len(self._plugin) == 0:
            raise ValueError(
                "No plugins loaded for this entry: %s\n"
                "A listing of installable plugins can be found "
                "at https://intake.readthedocs.io/en/latest/plugin"
                "-directory.html ." % self._driver
            )
        elif isinstance(self._plugin, list):
            plugin = self._plugin[0]
        else:
            # dict
            if plugin is None:
                # default selection for dict
                plugin = list(self._plugin)[0]
            spec = self._driver[plugin]
            open_args.update(spec.get("args", {}))
            try:
                plugin = self._plugin[plugin]
            except KeyError:
                raise ValueError("Attempt to select unavailable plugin %s, " "perhaps import of plugin failed" % plugin)
        return plugin, open_args

    def get(self, **user_parameters):
        """Instantiate the DataSource for the given parameters"""
        if not user_parameters and self._default_source is not None:
            return self._default_source

        plugin, open_args = self._create_open_args(user_parameters)
        data_source = plugin(**open_args)
        data_source.catalog_object = self._catalog
        data_source.name = self.name
        data_source.description = self._description
        data_source.cat = self._catalog

        # Cache the default source if there are no user parameters.
        if not user_parameters:
            self._default_source = data_source

        return data_source

    def clear_cached_default_source(self):
        """
        Clear a cached default source so it can be created anew (if, for instance,
        it depends on changing environment variables or execution context)
        """
        self._default_source = None


class CatalogParser(object):
    """Loads entries from a YAML spec"""

    def __init__(self, data, getenv=True, getshell=True, context=None):
        self._context = context if context else {}
        self._errors = []
        self._warnings = []
        self.getenv = getenv
        self.getshell = getshell
        self._data = self._parse(data)

    @property
    def ok(self):
        return len(self._errors) == 0

    @property
    def data(self):
        return self._data

    @property
    def errors(self):
        return self._errors

    @property
    def warnings(self):
        return self._warnings

    def error(self, msg, obj, key=None):
        if key is not None:
            self._errors.append(str((msg, obj, key)))
        else:
            self._errors.append(str((msg, obj)))

    def warning(self, msg, obj, key=None):
        if key is None:
            self._warnings.append(str((msg, obj)))
        else:
            self._warnings.append(str((msg, obj, key)))

    def _parse_plugins(self, data):
        if "plugins" not in data:
            return

        if not isinstance(data["plugins"], dict):
            self.error("value of key 'plugins' must be a dictionary", data, "plugins")
            return

        if "source" not in data["plugins"]:
            self.error("missing key 'source'", data["plugins"])
            return

        if not isinstance(data["plugins"]["source"], list):
            self.error("value of key 'source' must be a list", data["plugins"], "source")
            return

        for plugin_source in data["plugins"]["source"]:
            if not isinstance(plugin_source, dict):
                self.error("value in list of plugins sources must be a " "dictionary", data["plugins"], "source")
                continue

            if "module" in plugin_source:
                register_plugin_module(plugin_source["module"])
            elif "dir" in plugin_source:
                self.error(
                    "The key 'dir', and in general the feature of registering "
                    "plugins from a directory of Python scripts outside of "
                    "sys.path, is no longer supported. Use 'module'.",
                    plugin_source,
                )
            else:
                self.error("missing 'module'", plugin_source)

    def _getitem(self, obj, key, dtype, required=True, default=None, choices=None):
        if key in obj:
            if isinstance(obj[key], dtype):
                if choices and obj[key] not in choices:
                    self.error("value '{}' is invalid (choose from {})".format(obj[key], choices), obj, key)
                else:
                    return obj[key]
            else:
                self.error("value '{}' is not expected type '{}'".format(obj[key], dtype.__name__), obj, key)
            return None
        elif required:
            self.error("missing required key '{}'".format(key), obj)
            return None
        elif default:
            return default

        return None if dtype is object else dtype()

    def _parse_user_parameter(self, name, data):
        valid_types = list(COERCION_RULES)

        params = {
            "name": name,
            "description": self._getitem(data, "description", str),
            "type": self._getitem(data, "type", str, choices=valid_types),
            "default": self._getitem(data, "default", object, required=False),
            "min": self._getitem(data, "min", object, required=False),
            "max": self._getitem(data, "max", object, required=False),
            "allowed": self._getitem(data, "allowed", object, required=False),
        }

        if params["description"] is None or params["type"] is None:
            return None

        return UserParameter(**params)

    def _parse_data_source(self, name, data):
        if data.pop("remote", False):
            from intake.catalog.remote import RemoteCatalogEntry

            return RemoteCatalogEntry(getenv=self.getenv, getshell=self.getshell, **data)
        elif "cls" in data:
            from intake.utils import remake_instance

            return remake_instance(data)
        else:
            return self._parse_data_source_local(name, data)

    def _parse_data_source_local(self, name, data):
        ds = {
            "name": name,
            "description": self._getitem(data, "description", str, required=False),
            "driver": self._getitem(data, "driver", object),
            "direct_access": self._getitem(data, "direct_access", str, required=False, default="forbid", choices=["forbid", "allow", "force"]),
            "args": self._getitem(data, "args", dict, required=False),
            "cache": self._getitem(data, "cache", list, required=False),
            "metadata": self._getitem(data, "metadata", dict, required=False),
        }

        if ds["driver"] is None:
            return None

        ds["parameters"] = []

        if "parameters" in data:
            if isinstance(data["parameters"], list):
                raise exceptions.ObsoleteParameterError

            if not isinstance(data["parameters"], dict):
                self.error("value of key 'parameters' must be a dictionary", data, "parameters")
                return None

            for name, parameter in data["parameters"].items():
                if not isinstance(name, str):
                    self.error("key '{}' must be a string".format(name), data["parameters"], name)
                    continue

                if not isinstance(parameter, dict):
                    self.error("value of key '{}' must be a dictionary" "".format(name), data["parameters"], name)
                    continue

                obj = self._parse_user_parameter(name, parameter)
                if obj:
                    ds["parameters"].append(obj)

        return LocalCatalogEntry(catalog_dir=self._context["root"], getenv=self.getenv, getshell=self.getshell, **ds)

    def _parse_data_sources(self, data):
        sources = []

        if "sources" not in data:
            self.error("missing key 'sources'", data)
            return sources

        if isinstance(data["sources"], list):
            raise exceptions.ObsoleteDataSourceError

        if not isinstance(data["sources"], dict):
            self.error("value of key 'sources' must be a dictionary", data, "sources")
            return sources

        for name, source in data["sources"].items():
            if not isinstance(name, str):
                self.error("key '{}' must be a string".format(name), data["sources"], name)
                continue

            if not isinstance(source, dict):
                self.error("value of key '{}' must be a dictionary" "".format(name), data["sources"], name)
                continue

            obj = self._parse_data_source(name, source)
            if obj:
                sources.append(obj)

        return sources

    def _parse(self, data):
        if not isinstance(data, dict):
            self.error("catalog must be a dictionary", data)
            return

        return dict(
            plugin_sources=self._parse_plugins(data),
            data_sources=self._parse_data_sources(data),
            metadata=data.get("metadata", {}),
            name=data.get("name"),
            description=data.get("description"),
        )


def register_plugin_module(mod):
    """Find plugins in given module"""
    for k, v in load_plugins_from_module(mod).items():
        if k:
            if isinstance(k, (list, tuple)):
                k = k[0]
            # we clobber by default here
            register_driver(k, v, clobber=True)


def get_dir(path):
    if "://" in path:
        protocol, _ = split_protocol(path)
        out = get_filesystem_class(protocol)._parent(path)
        if "://" not in out:
            # some FSs strip this, some do not
            out = protocol + "://" + out
        return out
    path = make_path_posix(os.path.join(os.getcwd(), os.path.dirname(path)))
    if path[-1] != "/":
        path += "/"
    return path


class YAMLFileCatalog(Catalog):
    """Catalog as described by a single YAML file"""

    version = __version__
    container = "catalog"
    partition_access = None
    name = "yaml_file_cat"

    def __init__(self, path=None, text=None, autoreload=True, **kwargs):
        """
        Parameters
        ----------
        path: str
            Location of the file to parse (can be remote)
        text: str (DEPRECATED)
            YAML contents of catalog, takes precedence over path
        autoreload : bool
            Whether to watch the source file for changes; make False if you want
            an editable Catalog
        """
        self.path = path
        if text is not None:
            logger.warning("YAMLFileCatalog `text` argument is deprecated")
            warnings.warn("`text` argument is deprecated", DeprecationWarning)
            self.parse(text)
            self.autoreload = False
        else:
            self.autoreload = autoreload  # set this to False if don't want reloads
        self.filesystem = kwargs.pop("fs", None)
        self.access = "name" not in kwargs
        super(YAMLFileCatalog, self).__init__(**kwargs)

    def _load(self, reload=False):
        """Load text of catalog file and pass to parse

        Will do nothing if auto-reload is off and reload is not explicitly
        requested
        """
        if self.access is False:
            # skip first load, if cat has given name (i.e., is subcat)
            self.updated = 0
            self.access = True
            return
        if self.autoreload or reload:
            # First, we load from YAML, failing if syntax errors are found
            options = self.storage_options or {}
            if hasattr(self.path, "path") or hasattr(self.path, "read"):
                file_open = self.path
                self.path = make_path_posix(getattr(self.path, "path", getattr(self.path, "name", "file")))
            elif self.filesystem is None:
                file_open = open_files(self.path, mode="rb", **options)
                assert len(file_open) == 1
                file_open = file_open[0]
                self.filesystem = file_open.fs
            else:
                file_open = self.filesystem.open(self.path, mode="rb")
            self._dir = get_dir(self.path)

            with file_open as f:
                text = f.read().decode()
            if "!template " in text:
                logger.warning("Use of '!template' deprecated - fixing")
                text = text.replace("!template ", "")
            self.parse(text)

    def add(self, source, name=None, path=None, storage_options=None):
        """Add sources to the catalog and save into the original file

        This adds the source into the catalog dictionary, and saves the
        resulting catalog as YAML. Typically, this would be used to update a
        catalog file in-place. Optionally, the new catalog can be saved to a
        new location, in which case the new catalog is returned.

        Note that if a source of the given name exists, it will be clobbered.

        Parameters
        ----------
        source : DataSource instance
            The source whose spec we want to save
        name : str or None
            The name the source is to have in the catalog; use the source's
            name attribute, if not given.
        path : str or None
            Location to save the new catalog; if None, the original location
            from which it was loaded
        storage_options : dict or None
            If saving to a new location, use these arguments for the filesystem
            backend

        Returns
        -------
        YAMLFileCatalog instance, containing the new entry
        """
        import yaml

        entries = self._entries.copy()
        name = name or source.name or "source"
        entries[name] = source

        if path is None:
            options = self.storage_options or {}
            file_open = open_files([self.path], mode="wt", **options)
        else:
            options = storage_options or {}
            file_open = open_files([path], mode="wt", **options)
        assert len(file_open) == 1
        file_open = file_open[0]

        data = {"metadata": self.metadata, "sources": {}}
        for e in entries:
            data["sources"][e] = list(entries[e]._yaml()["sources"].values())[0]
        with file_open as f:
            yaml.dump(data, f, default_flow_style=False)

        if path:
            return self
        else:
            return YAMLFileCatalog(self.path, storage_options=storage_options, autoreload=self.autoreload)

    def parse(self, text):
        """Create entries from catalog text

        Normally the text comes from the file at self.path via the ``_load()``
        method, but could be explicitly set instead. A copy of the text is
        kept in attribute ``.text`` .

        Parameters
        ----------
        text : str
            YAML formatted catalog spec
        """
        self.text = text
        data = yaml_load(self.text)

        if data is None:
            raise exceptions.CatalogException("No YAML data in file")

        # Second, we validate the schema and semantics
        context = dict(root=self._dir)
        result = CatalogParser(data, context=context, getenv=self.getenv, getshell=self.getshell)
        if result.errors:
            raise exceptions.ValidationError("Catalog '{}' has validation errors:\n\n{}" "".format(self.path, "\n".join(result.errors)), result.errors)

        cfg = result.data

        self._entries = {}
        shared_parameters = data.get("metadata", {}).get("parameters", {})
        self.user_parameters.update({name: UserParameter(name, **attrs) for name, attrs in shared_parameters.items()})

        for entry in cfg["data_sources"]:
            entry._catalog = self
            self._entries[entry.name] = entry
            entry._filesystem = self.filesystem

        meta = self.metadata.copy()
        meta.update(cfg.get("metadata", {}))
        self.metadata = meta
        self.name = self.name or cfg.get("name") or self.name_from_path
        self.description = self.description or cfg.get("description")

    @property
    def name_from_path(self):
        """If catalog is named 'catalog' take name from parent directory"""
        name = os.path.splitext(os.path.basename(self.path))[0]
        if name == "catalog":
            name = os.path.basename(os.path.dirname(self.path))
        return name.replace(".", "_")


class YAMLFilesCatalog(Catalog):
    """Catalog as described by a multiple YAML files"""

    version = (__version__,)
    container = "catalog"
    partition_access = None
    name = "yaml_files_cat"

    def __init__(self, path, flatten=True, **kwargs):
        """
        Parameters
        ----------
        path: str
            Location of the files to parse (can be remote), including possible
            glob (*) character(s). Can also be list of paths, without glob
            characters.
        flatten: bool (True)
            Whether to list all entries in the cats at the top level (True)
            or create sub-cats from each file (False).
        """
        self.path = path
        self._flatten = flatten
        self._kwargs = kwargs.copy()
        self._cat_files = []
        self._cats = {}
        self.access = "name" not in kwargs
        super(YAMLFilesCatalog, self).__init__(**kwargs)

    def _load(self):
        # initial: find cat files
        # if flattening, need to get all entries from each.
        if self.access is False:
            # skip first load, if cat has given name (i.e., is subcat)
            self.updated = 0
            self.access = True
            return
        self._entries.clear()
        options = self.storage_options or {}
        if isinstance(self.path, (list, tuple)):
            files = sum([open_files(p, mode="rb", **options) for p in self.path], [])
            self.name = self.name or "%i files" % len(files)
            self.description = self.description or f"Catalog generated from {len(files)} files"
            self.path = [make_path_posix(p) for p in self.path]
        else:
            if isinstance(self.path, str) and "*" not in self.path:
                self.path = self.path + "/*"
            files = open_files(self.path, mode="rb", **options)
            self.path = make_path_posix(self.path)
            self.name = self.name or self.path
            self.description = self.description or f"Catalog generated from all files found in {self.path}"
        if not set(f.path for f in files) == set(f.path for f in self._cat_files):
            # glob changed, reload all
            self._cat_files = files
            self._cats.clear()
        for f in files:
            name = os.path.split(f.path)[-1].replace(".yaml", "").replace(".yml", "")
            kwargs = self.kwargs.copy()
            kwargs["path"] = f.path
            d = make_path_posix(os.path.dirname(f.path))
            if f.path not in self._cats:
                entry = LocalCatalogEntry(name, "YAML file: %s" % name, "yaml_file_cat", True, kwargs, [], [], self.metadata, d)
                if self._flatten:
                    # store a concrete Catalog
                    try:
                        cat = entry()
                        cat.reload()
                        self.user_parameters.update(cat.user_parameters)
                        self._cats[f.path] = cat
                    except IOError as e:
                        logger.info('Loading "%s" as a catalog failed: %s' "" % (entry, e))
                else:
                    # store a catalog entry
                    self._cats[f.path] = entry
        entries = {}
        for name, entry in list(self._cats.items()):
            if self._flatten:
                entry.reload()
                inter = set(entry._entries).intersection(entries)
                if inter:
                    raise ValueError("Conflicting names when flattening multiple" " catalogs. Sources %s exist in more than" " one" % inter)
                entries.update(entry._entries)
            else:
                entries[entry._name] = entry
        self._entries.update(entries)


class MergedCatalog(Catalog):
    """
    A Catalog that merges the entries of a list of catalogs.
    """

    def __init__(self, catalogs, *args, **kwargs):
        self._catalogs = catalogs
        super().__init__(*args, **kwargs)

    def _load(self):
        for catalog in self._catalogs:
            catalog._load()
        self._entries = collections.ChainMap(*(catalog._entries for catalog in self._catalogs))


class EntrypointEntry(CatalogEntry):
    """
    A catalog entry for an entrypoint.

    """

    def __init__(self, entrypoint):
        self._entrypoint = entrypoint
        self._container = None
        self._user_parameters = []
        super().__init__()

    def __repr__(self):
        return f"<Entry '{self.name}'>"

    @property
    def name(self):
        return self._entrypoint.name

    def describe(self):
        """Basic information about this entry"""
        if self._container is None:
            self._container = self().container
        return {
            "name": self.name,
            "module_name": self._entrypoint.module_name,
            "object_name": self._entrypoint.object_name,
            "distro": self._entrypoint.distro,
            "extras": self._entrypoint.extras,
            "container": self._container,
        }

    def get(self):
        """Instantiate the DataSource for the given parameters"""
        return self._entrypoint.load()


class EntrypointsCatalog(Catalog):
    """
    A catalog of discovered entrypoint catalogs.
    """

    def __init__(self, *args, entrypoints_group="intake.catalogs", paths=None, **kwargs):
        self._entrypoints_group = entrypoints_group
        self._paths = paths
        super().__init__(*args, **kwargs)

    def _load(self):
        catalogs = entrypoints.get_group_named(self._entrypoints_group, path=self._paths)
        self.name = self.name or "EntrypointsCatalog"
        self.description = self.description or f"EntrypointsCatalog of {len(catalogs)} catalogs."
        for name, entrypoint in catalogs.items():
            try:
                self._entries[name] = EntrypointEntry(entrypoint)
            except Exception as e:
                warnings.warn(f"Failed to load {name}, {entrypoint}, {e!r}.")


# Register these early in the import process to support the default catalog
# which is built at import time. (Without this, 'yaml_file_cat' is looked for
# in intake.registry before the registry has been populated.)
register_driver("yaml_file_cat", YAMLFileCatalog, clobber=True)
register_driver("yaml_files_cat", YAMLFilesCatalog, clobber=True)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import logging
import os

import yaml

from intake.cli.util import Subcommand

log = logging.getLogger(__name__)


class Cache(Subcommand):
    """Locally cached files"""

    name = "cache"

    def initialize(self):
        sub_parser = self.parser.add_subparsers()

        cache_list = sub_parser.add_parser("list-keys", help="List keys currently stored")
        cache_list.set_defaults(invoke=self._list_keys)

        cache_files = sub_parser.add_parser("list-files", help="List files for a give key")
        cache_files.add_argument("key", type=str, help="Key to list files for")
        cache_files.set_defaults(invoke=self._list_files)

        cache_rm = sub_parser.add_parser("clear", help="Clear a key from the cache")
        cache_rm.add_argument("key", type=str, help="Key to remove (all, if omitted)", nargs="?")
        cache_rm.set_defaults(invoke=self._clear)

        cache_du = sub_parser.add_parser("usage", help="Print usage information")
        cache_du.set_defaults(invoke=self._usage)

    def invoke(self, args):
        self.parser.print_help()

    def _clear(self, args):
        from intake.source.cache import BaseCache

        if args.key is None:
            BaseCache(None, None).clear_all()
        else:
            BaseCache(None, None).clear_cache(args.key)

    def _list_keys(self, args):
        from intake.source.cache import CacheMetadata

        md = CacheMetadata()
        print(yaml.dump(list(md), default_flow_style=False))

    def _list_files(self, args):
        from intake.source.cache import CacheMetadata

        md = CacheMetadata()
        print(yaml.dump(md[args.key], default_flow_style=False))

    def _usage(self, args):
        from intake.config import conf

        total_size = 0
        for dirpath, dirnames, filenames in os.walk(os.path.join(conf["cache_dir"], "cache")):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)
        for unit in ["", "k", "M", "G", "T", "P", "E", "Z"]:
            # "human"
            # https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597
            if abs(total_size) < 1024.0:
                s = "%3.1f %s" % (total_size, unit)
                break
            total_size /= 1024.0
        print("%s: %s" % (conf["cache_dir"], s))



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
"""

"""
import logging

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import os
import shutil

# Standard library imports
from os.path import dirname, exists, join

# Intake imports
from intake.cli.util import Subcommand

# External imports


# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


class Example(Subcommand):
    """Create example catalog"""

    name = "example"

    def initialize(self):
        pass

    def invoke(self, args):
        print("Creating example catalog...")
        files = ["us_states.yml", "states_1.csv", "states_2.csv"]
        for filename in files:
            if exists(filename):
                print("Cannot create example catalog in current directory.\n" "%s already exists." % filename)
                return 1

        src_dir = join(dirname(__file__), "..", "..", "sample")

        for filename in files:
            src_name = join(src_dir, filename)
            dest_name = filename
            dest_dir = dirname(filename)
            print("  Writing %s" % filename)
            if dest_dir != "" and not exists(dest_dir):
                os.mkdir(dest_dir)
            shutil.copyfile(src_name, dest_name)

        print(
            """\nTo load the catalog:
    >>> import intake
    >>> cat = intake.open_catalog('%s')
    """
            % files[0]
        )



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
from functools import partial

import panel as pn

from ..base import MAX_WIDTH, Base, enable_widget
from .add import CatAdder
from .search import Search
from .select import CatSelector


class CatGUI(Base):
    """
    Top level GUI panel that contains controls and all visible sub-panels

    This class is responsible for coordinating the inputs and outputs
    of various sup-panels and their effects on each other.

    Parameters
    ----------
    cats: list of catalogs
        catalogs used to initalize the select
    done_callback: func, opt
        called when the object's main job has completed. In this case,
        selecting catalog(s).

    Attributes
    ----------
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """

    def __init__(self, cats=None, done_callback=None, **kwargs):
        self._cats = cats
        self.panel = pn.Column(name="Catalogs", width_policy="max", max_width=MAX_WIDTH)
        self.done_callback = done_callback or (lambda x: x)

        self.add_widget = pn.widgets.Toggle(name="＋", value=False, disabled=False, width=50)

        self.remove_widget = pn.widgets.Button(name="―", disabled=True, width=50)

        self.search_widget = pn.widgets.Toggle(name="🔍", value=False, disabled=True, width=50)

        self.controls = [
            self.add_widget,
            self.remove_widget,
            self.search_widget,
        ]
        self.control_panel = pn.Row(name="Controls", margin=0)

        self.select = CatSelector(cats=self._cats, done_callback=self.callback)
        self.add = CatAdder(done_callback=self.select.add, visible=self.add_widget.value, visible_callback=partial(setattr, self.add_widget, "value"))
        self.search = Search(
            cats=self.cats, done_callback=self.select.add, visible=self.search_widget.value, visible_callback=partial(setattr, self.search_widget, "value")
        )

        self.children = [
            self.select.panel,
            self.control_panel,
            self.search.panel,
            self.add.panel,
        ]

        super().__init__(**kwargs)

    def setup(self):
        self.watchers = [
            self.add_widget.link(self.add, value="visible"),
            self.search_widget.param.watch(self.on_click_search_widget, "value"),
            self.remove_widget.param.watch(self.select.remove_selected, "clicks"),
        ]

    @Base.visible.setter
    def visible(self, visible):
        """When visible changed, do setup or unwatch and call visible_callback"""
        self._visible = visible

        if visible and len(self._panel.objects) == 0:
            self.setup()
            self.select.visible = True
            self.control_panel.extend(self.controls)
            self._panel.extend(self.children)
        elif not visible and len(self._panel.objects) > 0:
            self.unwatch()
            # do children
            self.select.visible = False
            self.control_panel.clear()
            self.search.visible = False
            self.add.visible = False
            self._panel.clear()
        if self.visible_callback:
            self.visible_callback(visible)

    def callback(self, cats):
        """When a catalog is selected, enable widgets that depend on that condition
        and do done_callback"""
        enable = bool(cats)
        if not enable:
            # close search if it is visible
            self.search.visible = False
        enable_widget(self.search_widget, enable)
        enable_widget(self.remove_widget, enable)

        if self.done_callback:
            self.done_callback(cats)

    def on_click_search_widget(self, event):
        """When the search control is toggled, set visibility and hand down cats"""
        self.search.cats = self.cats
        self.search.visible = event.new
        if self.search.visible:
            self.search.watchers.append(self.select.widget.link(self.search, value="cats"))

    @property
    def cats(self):
        """Cats that have been selected from the select"""
        return self.select.selected

    def __getstate__(self):
        """Serialize the current state of the object"""
        return {
            "visible": self.visible,
            "select": self.select.__getstate__(),
            "add": self.add.__getstate__(),
            "search": self.search.__getstate__(include_cats=False),
        }

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.visible = state.get("visible", True)
        if self.visible:
            self.select.__setstate__(state["select"])
            self.add.__setstate__(state["add"])
            self.search.__setstate__(state["search"])
        return self

    @classmethod
    def from_state(cls, state):
        """Create a new object from a serialized verion of another one.

        Example
        -------
        original = CatGUI()
        copy = CatGUI.from_state(original.__getstate__())
        """
        return cls(cats=[]).__setstate__(state)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import os
import shutil
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager

import requests
import yaml

from .utils import make_path_posix

ex = sys.executable
here = os.path.dirname(__file__)
defcat = make_path_posix(os.path.join(here, "cli", "server", "tests", "catalog1.yml"))
PY2 = sys.version_info[0] == 2


@contextmanager
def tempdir():
    d = tempfile.mkdtemp()
    try:
        yield d
    finally:
        if os.path.exists(d):
            shutil.rmtree(d)


@contextmanager
def temp_conf(conf):
    with tempdir() as d:
        fn = os.path.join(d, "conf.yaml")
        with open(fn, "w") as f:
            yaml.dump(conf, f)
        yield fn


@contextmanager
def server(args=None, cat=None, env=None, wait=None, timeout=25):
    cat = cat if cat is not None else defcat
    args = list(args if args is not None else []) + []
    env = env if env is not None else {}
    cmd = [ex, "-m", "intake.cli.server"] + list(args) + [cat]
    p = subprocess.Popen(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    if wait is not None:
        while True:
            try:
                requests.get("http://localhost:%i/v1/info" % wait)
                break
            except:
                time.sleep(0.1)
                timeout -= 0.1
                assert timeout > 0
    try:
        yield p
    finally:
        p.terminate()



from functools import lru_cache

from .. import open_catalog
from . import import_name
from .base import DataSource, Schema


class AliasSource(DataSource):
    """Refer to another named source, unmodified

    The purpose of an Alias is to be able to refer to other source(s) in the
    same catalog, perhaps leaving the choice of which target to load up to the
    user. This source makes no sense outside of a catalog.

    In this case, the output of the target source is not modified, but this
    class acts as a prototype 'derived' source for processing the output of
    some standard driver.

    After initial discovery, the source's container and other details will be
    updated from the target; initially, the AliasSource container is not
    any standard.
    """

    container = "other"
    version = 1
    name = "alias"

    def __init__(self, target, mapping=None, metadata=None, **kwargs):
        """

        Parameters
        ----------
        target: str
            Name of the source to load, must be a key in the same catalog
        mapping: dict or None
            If given, use this to map the string passed as ``target`` to
            entries in the catalog
        metadata: dict or None
            Extra metadata to associate
        kwargs: passed on to the target
        """
        super(AliasSource, self).__init__(metadata)
        self.target = target
        self.mapping = mapping or {target: target}
        self.kwargs = kwargs
        self.metadata = metadata
        self.source = None

    def _get_source(self):
        if self.cat is None:
            raise ValueError("AliasSource cannot be used outside a catalog")
        if self.source is None:
            self.source = self.cat[self.mapping[self.target]](metadata=self.metadata, **self.kwargs)
            self.metadata = self.source.metadata.copy()
            self.container = self.source.container
            self.partition_access = self.source.partition_access
            self.description = self.source.description

    def discover(self):
        self._get_source()
        return self.source.discover()

    def read(self):
        self._get_source()
        return self.source.read()

    def read_partition(self, i):
        self._get_source()
        return self.source.read_partition(i)

    def read_chunked(self):
        self._get_source()
        return self.source.read_chunked()

    def to_dask(self):
        self._get_source()
        return self.source.to_dask()


cached_cats = lru_cache(10)(open_catalog)


def get_source(target, cat, kwargs, cat_kwargs):
    if ":" in target:
        caturl, target = target.rsplit(":", 1)
        cat = cached_cats(caturl, **cat_kwargs)
    if cat:
        return cat[target].configure_new(**kwargs)
    # for testing only
    return target  # pragma: no cover


def first(targets, cat, kwargs, cat_kwargs):
    """A target chooser that simply picks the first from the given list

    This is the default, particularly for the case of only one element in
    the list
    """
    targ = targets[0]
    return get_source(targ, cat, kwargs.get(targ, {}), cat_kwargs)


def first_discoverable(targets, cat, kwargs, cat_kwargs):
    """A target chooser: the first target for which discover() succeeds

    This may be useful where some drivers are not importable, or some
    sources can be available only sometimes.
    """
    for t in targets:
        try:
            s = get_source(t, cat, kwargs.get(t, {}), cat_kwargs)
            s.discover()
            return s
        except Exception:
            pass
    raise RuntimeError("No targets succeeded at discover()")


class DerivedSource(DataSource):
    """Base source deriving from another source in the same catalog

    Target picking and parameter validation are performed here, but
    you probably want to subclass from one of the more specific
    classes like ``DataFrameTransform``.
    """

    input_container = "other"  # no constraint
    container = "other"  # to be filled in per instance at access time
    required_params = []  # list of kwargs that must be present
    optional_params = {}  # optional kwargs with defaults

    def __init__(self, targets, target_chooser=first, target_kwargs=None, cat_kwargs=None, container=None, metadata=None, **kwargs):
        """

        Parameters
        ----------
        targets: list of string or DataSources
            If string(s), refer to entries of the same catalog as this Source
        target_chooser: function to choose between targets
            function(targets, cat) -> source, or a fully-qualified dotted string pointing
            to it
        target_kwargs: dict of dict with keys matching items of targets
        cat_kwargs: to pass to intake.open_catalog, if the target is in
            another catalog
        container: str (optional)
            Assumed output container, if known/different from input

        [Note: the exact form of target_kwargs and cat_kwargs may be
        subject to change]
        """
        self.targets = targets
        self._chooser = target_chooser if callable(target_chooser) else import_name(target_chooser)
        self._kwargs = target_kwargs or {}
        self._source = None
        self._params = kwargs
        self._cat_kwargs = cat_kwargs or {}
        if container:
            self.container = container
        self._validate_params()
        super().__init__(metadata=metadata)

    def _validate_params(self):
        """That all required params are present and that optional types match"""
        assert set(self.required_params) - set(self._params) == set()
        for par, val in self.optional_params.items():
            if par not in self._params:
                self._params[par] = val

    def _pick(self):
        """Pick the source from the given targets"""
        self._source = self._chooser(self.targets, self.cat, self._kwargs, self._cat_kwargs)
        if self.input_container != "other":
            assert self._source.container == self.input_container

        self.metadata["target"] = self._source.metadata
        if self.container is None:
            self.container = self._source.container


class GenericTransform(DerivedSource):
    required_params = ["transform", "transform_kwargs"]
    optional_params = {"allow_dask": True}
    """
    Perform an arbitrary function to transform an input

        transform: function to perform transform
            function(container_object) -> output, or a fully-qualified dotted string pointing
            to it
        transform_params: dict
            The keys are names of kwargs to pass to the transform function. Values are either
            concrete values to pass; or param objects which can be made into widgets (but
            must have a default value) - or a spec to be able to make these objects.
        allow_dask: bool (optional, default True)
            Whether to_dask() is expected to work, which will in turn call the
            target's to_dask()
    """

    def _validate_params(self):
        super()._validate_params()
        transform = self._params["transform"]
        self._transform = transform if callable(transform) else import_name(transform)

    def _get_schema(self):
        """We do not know the schema of a generic transform"""
        self._pick()
        return Schema()

    def to_dask(self):
        self._get_schema()
        if not self._params["allow_dask"]:
            raise ValueError("This transform is not compatible with Dask" "because it has use_dask=False")
        return self._transform(self._source.to_dask(), **self._params["transform_kwargs"])

    def read(self):
        self._get_schema()
        return self._transform(self._source.read(), **self._params["transform_kwargs"])


class DataFrameTransform(GenericTransform):
    """Transform where the input and output are both Dask-compatible dataframes

    This derives from GenericTransform, and you must supply ``transform`` and
    any ``transform_kwargs``.
    """

    input_container = "dataframe"
    container = "dataframe"
    optional_params = {}
    _df = None

    def to_dask(self):
        if self._df is None:
            self._pick()
            self._df = self._transform(self._source.to_dask(), **self._params["transform_kwargs"])
        return self._df

    def _get_schema(self):
        """load metadata only if needed"""
        self.to_dask()
        return Schema(dtype=self._df.dtypes, shape=(None, len(self._df.columns)), npartitions=self._df.npartitions, metadata=self.metadata)

    def read(self):
        return self.to_dask().compute()


class Columns(DataFrameTransform):
    """Simple dataframe transform to pick columns

    Given as an example of how to make a specific dataframe transform.
    Note that you could use DataFrameTransform directly, by writing a
    function to choose the columns instead of a method as here.
    """

    input_container = "dataframe"
    container = "dataframe"
    required_params = ["columns"]

    def __init__(self, columns, **kwargs):
        """
        columns: list of labels (usually str) or slice
            Columns to choose from the target dataframe
        """
        # this class wants requires "columns", but DataFrameTransform
        # uses "transform_kwargs", which we don't need since we use a method for the
        # transform
        kwargs.update(transform=self.pick_columns, columns=columns, transform_kwargs={})
        super().__init__(**kwargs)

    def pick_columns(self, df):
        return df[self._params["columns"]]



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
""" Provide a ``main`` function to run intake commands.

"""

import logging

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

# Standard library imports
import argparse

# Intake imports
from intake import __version__
from intake.cli.util import die, nice_join

# External imports


# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


def main(description, subcommands, argv):
    """Execute an intake command.

    Args:
        description (str) :
            A description for this top-level command

        subcommands (seq[SubCommand]) :
            A list of subcommands to configure for argparse

        argv (seq[str]) :
            A list of command line arguments to process

    Returns:
        None

    """
    if len(argv) == 1:
        die("ERROR: Must specify subcommand, one of: %s" % nice_join(x.name for x in subcommands))

    parser = argparse.ArgumentParser(prog=argv[0], description=description, epilog="See '<command> --help' to read about a specific subcommand.")

    parser.add_argument("-v", "--version", action="version", version=__version__)

    subs = parser.add_subparsers(help="Sub-commands")

    for cls in subcommands:
        subparser = subs.add_parser(cls.name, help=cls.__doc__.strip())
        subcommand = cls(parser=subparser)
        subparser.set_defaults(invoke=subcommand.invoke)

    args = parser.parse_args(argv[1:])
    try:
        return args.invoke(args) or 0  # convert None to 0
    except Exception as e:
        die("ERROR: " + repr(e))



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import msgpack
import requests
from requests.compat import urljoin

from intake import __version__
from intake.source.base import DataSource, Schema

from ..compat import pack_kwargs, unpack_kwargs
from . import serializer


class RemoteSource(DataSource):
    """Base class for all DataSources living on an Intake server"""

    version = __version__

    def __init__(self, url, headers, name, parameters, metadata=None, **kwargs):
        """

        Parameters
        ----------
        url: str
            Address of the server
        headers: dict
            HTTP headers to sue in calls
        name: str
            handle to reference this data
        parameters: dict
            To pass to the server when it instantiates the data source
        metadata: dict
            Additional info
        kwargs: ignored
        """
        super(RemoteSource, self).__init__(self)
        self.url = url
        self.name = name
        self.parameters = parameters
        self.headers = headers
        self._source_id = None
        self.metadata = metadata or {}
        self._get_source_id()

    def _get_source_id(self):
        if self._source_id is None:
            payload = dict(action="open", name=self.name, parameters=self.parameters)
            req = requests.post(urljoin(self.url, "v1/source"), data=msgpack.packb(payload, **pack_kwargs), **self.headers)
            req.raise_for_status()
            response = msgpack.unpackb(req.content, **unpack_kwargs)
            self._parse_open_response(response)

    def _parse_open_response(self, response):
        dtype_descr = response["dtype"]
        if isinstance(dtype_descr, list):
            # Reformat because NumPy needs list of tuples
            dtype_descr = [tuple(x) for x in response["dtype"]]
        self.dtype = dtype_descr
        self.shape = tuple(response["shape"] or ())
        self.npartitions = response["npartitions"]
        self.metadata = response["metadata"]
        self._schema = Schema(datashape=None, dtype=self.dtype, shape=self.shape, npartitions=self.npartitions, metadata=self.metadata)
        self._source_id = response["source_id"]

    def _get_partition(self, i):
        raise NotImplementedError

    def to_dask(self):
        raise NotImplementedError


def get_partition(url, headers, source_id, container, partition):
    """Serializable function for fetching a data source partition

    Parameters
    ----------
    url: str
        Server address
    headers: dict
        HTTP header parameters
    source_id: str
        ID of the source in the server's cache (unique per user)
    container: str
        Type of data, like "dataframe" one of ``intake.container.container_map``
    partition: serializable
        Part of data to fetch, e.g., an integer for a dataframe.
    """
    accepted_formats = list(serializer.format_registry.keys())
    accepted_compression = list(serializer.compression_registry.keys())
    payload = dict(action="read", source_id=source_id, accepted_formats=accepted_formats, accepted_compression=accepted_compression)

    if partition is not None:
        payload["partition"] = partition

    try:
        resp = requests.post(urljoin(url, "v1/source"), data=msgpack.packb(payload, **pack_kwargs), **headers)
        if resp.status_code != 200:
            raise Exception("Error reading data")

        msg = msgpack.unpackb(resp.content, **unpack_kwargs)
        format = msg["format"]
        compression = msg["compression"]
        compressor = serializer.compression_registry[compression]
        encoder = serializer.format_registry[format]
        chunk = encoder.decode(compressor.decompress(msg["data"]), container)
        return chunk
    finally:
        if resp is not None:
            resp.close()



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
"""
"""

import logging

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

# Standard library imports

# External imports

# Intake imports
from intake import open_catalog
from intake.cli.util import Subcommand

# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


class Precache(Subcommand):
    """Populate caching for catalog entries that define caching."""

    name = "precache"

    def initialize(self):
        self.parser.add_argument("uri", metavar="URI", type=str, help="Catalog URI")

    def invoke(self, args):
        catalog = open_catalog(args.uri)
        for entry in list(catalog):
            try:
                s = catalog[entry]
                s.read()
                if s.cache:
                    print("Caching for entry %s" % entry)
            except Exception as e:
                print("Skipping {} due to {}".format(entry, e))



from .base import Catalog
from .local import LocalCatalogEntry


class ZarrGroupCatalog(Catalog):
    """A catalog of the members of a Zarr group."""

    version = "0.0.1"
    container = "catalog"
    partition_access = None
    name = "zarr_cat"

    def __init__(self, urlpath, storage_options=None, component=None, metadata=None, consolidated=False, name=None):
        """

        Parameters
        ----------
        urlpath : str
            Location of data file(s), possibly including protocol information
        storage_options : dict, optional
            Passed on to storage backend for remote files
        component : str, optional
            If None, build a catalog from the root group. If given, build the
            catalog from the group at this location in the hierarchy.
        metadata : dict, optional
            Catalog metadata. If not provided, will be populated from Zarr
            group attributes.
        consolidated : bool, optional
            If True, assume Zarr metadata has been consolidated.
        """
        self._urlpath = urlpath
        self._storage_options = storage_options or {}
        self._component = component
        self._consolidated = consolidated
        self._grp = None
        self.name = name
        super().__init__(metadata=metadata)

    def _load(self):
        import zarr

        if self._grp is None:
            # obtain the zarr root group
            if isinstance(self._urlpath, zarr.hierarchy.Group):
                # use already-opened group, allows support for nested groups
                # as catalogs
                root = self._urlpath

            else:
                # obtain store
                if isinstance(self._urlpath, str):
                    # open store from url
                    from fsspec import get_mapper

                    store = get_mapper(self._urlpath, **self._storage_options)
                else:
                    # assume store passed directly
                    store = self._urlpath

                # open root group
                if self._consolidated:
                    # use consolidated metadata
                    root = zarr.open_consolidated(store=store, mode="r")
                else:
                    root = zarr.open_group(store=store, mode="r")

            # deal with component path
            if self._component is None:
                self._grp = root
            else:
                self._grp = root[self._component]

            # use zarr attributes as metadata
            self.metadata.update(self._grp.attrs.asdict())

        # build catalog entries
        entries = {}
        for k, v in self._grp.items():
            if isinstance(v, zarr.core.Array):
                entry = LocalCatalogEntry(name=k, description="", driver="ndzarr", args=dict(urlpath=v), catalog=self)
            else:
                entry = LocalCatalogEntry(name=k, description="", driver="zarr_cat", args=dict(urlpath=v))
            entries[k] = entry
        self._entries = entries

    def to_zarr(self):
        return self._grp
